import numpy as np
import pytest

import npx


@pytest.mark.parametrize(
    ("p", "ref"),
    [
        (-np.inf, 1.0),  # min
        (-20000, 1.0000693171203765),
        (-1, 1.9672131147540985),  # harmonic mean
        (-0.1, 2.3000150292740735),
        (0, 2.3403473193207156),  # geometric mean
        (0.1, 2.3810581190184337),
        (1, 2.75),  # arithmetic mean
        (2, np.sqrt(9.75)),  # root mean square
        (10000, 4.999306900862521),
        (np.inf, 5.0),  # max
    ],
)
def test_mean_pos(p, ref):
    a = [1.0, 2.0, 3.0, 5.0]
    val = npx.mean(a, p)
    # val = pmean(a, p)
    # print(p, val)
    assert abs(val - ref) < 1.0e-13 * abs(ref)


@pytest.mark.parametrize(
    ("p", "ref"),
    [
        (-np.inf, 1.0),  # absmin
        (-1, -1.9672131147540985),  # harmonic mean
        # (0, 2.3403473193207156), # geometric mean
        (1, -2.75),  # arithmetic mean
        (2, np.sqrt(9.75)),  # root mean square
        (np.inf, 5.0),  # absmax
    ],
)
def test_mean_neg(p, ref):
    a = [-1.0, -2.0, -3.0, -5.0]
    val = npx.mean(a, p)
    # val = pmean(a, p)
    # print(p, val)
    assert abs(val - ref) < 1.0e-13 * abs(ref)


def test_errors():
    a = [-1.0, -2.0, -3.0, -5.0]
    with pytest.raises(TypeError, match="Non-integer p.*"):
        npx.mean(a, 0.5)

    with pytest.raises(ValueError, match="p=0 only works with nonnegative x."):
        npx.mean(a, 0)



from functools import reduce
from operator import mul

import numpy as np
from numpy.typing import ArrayLike


# math.prod in 3.8
# https://docs.python.org/3/library/math.html#math.prod
def _prod(a):
    return reduce(mul, a, 1)


def dot(a: ArrayLike, b: np.ndarray) -> np.ndarray:
    """Take arrays `a` and `b` and form the dot product between the last axis
    of `a` and the first of `b`.
    """
    return np.tensordot(a, b, 1)


def outer(a: ArrayLike, b: ArrayLike) -> np.ndarray:
    """Compute the outer product of two arrays `a` and `b` such that the shape
    of the resulting array is `(*a.shape, *b.shape)`.
    """
    a = np.asarray(a)
    b = np.asarray(b)
    return np.outer(a, b).reshape(*a.shape, *b.shape)


def solve(A: np.ndarray, x: np.ndarray) -> np.ndarray:
    """Solves a linear equation system with a matrix of shape (n, n) and an array of
    shape (n, ...). The output has the same shape as the second argument.
    """
    # https://stackoverflow.com/a/48387507/353337
    x = np.asarray(x)
    return np.linalg.solve(A, x.reshape(x.shape[0], -1)).reshape(x.shape)


def sum_at(a: ArrayLike, indices: ArrayLike, minlength: int):
    """Sums up values `a` with `indices` into an output array of at least
    length `minlength` while treating dimensionality correctly. It's a lot
    faster than numpy's own np.add.at (see
    https://github.com/numpy/numpy/issues/5922#issuecomment-511477435).

    Typically, `indices` will be a one-dimensional array; `a` can have any
    dimensionality. In this case, the output array will have shape (minlength,
    a.shape[1:]).

    `indices` may have arbitrary shape, too, but then `a` has to start out the
    same. (Those dimensions are flattened out in the computation.)
    """
    a = np.asarray(a)
    indices = np.asarray(indices)

    if len(a.shape) < len(indices.shape):
        msg = (
            f"a.shape = {a.shape}, indices.shape = {indices.shape}, "
            "but len(a.shape) >= len(indices.shape) is required."
        )
        raise RuntimeError(msg)

    m = len(indices.shape)
    assert indices.shape == a.shape[:m]

    out_shape = (minlength, *a.shape[m:])

    indices = indices.reshape(-1)
    a = a.reshape(_prod(a.shape[:m]), _prod(a.shape[m:]))

    # Cast to int; bincount doesn't work for uint64 yet
    # https://github.com/numpy/numpy/issues/17760
    indices = indices.astype(int)

    return np.array(
        [
            np.bincount(indices, weights=a[:, k], minlength=minlength)
            for k in range(a.shape[1])
        ],
    ).T.reshape(out_shape)


def add_at(a: ArrayLike, indices: ArrayLike, b: ArrayLike):
    a = np.asarray(a)
    indices = np.asarray(indices)
    b = np.asarray(b)

    m = len(indices.shape)
    assert a.shape[1:] == b.shape[m:]
    a += sum_at(b, indices, a.shape[0])


def subtract_at(a: ArrayLike, indices: ArrayLike, b: ArrayLike):
    b = np.asarray(b)
    add_at(a, indices, -b)



import numpy as np

import npx

rng = np.random.default_rng(0)


def test_dot():
    a = rng.random((1, 2, 3))
    b = rng.random((3, 4, 5))
    c = npx.dot(a, b)
    assert c.shape == (1, 2, 4, 5)


def test_solve():
    a = rng.random((3, 3))
    b = rng.random((3, 4, 5))
    c = npx.solve(a, b)
    assert c.shape == b.shape


def test_outer():
    a = rng.random((1, 2))
    b = rng.random((3, 4))
    c = npx.outer(a, b)
    assert c.shape == (1, 2, 3, 4)



import numpy as np
from numpy.typing import ArrayLike


# There also is
# <https://docs.scipy.org/doc/scipy/reference/generated/scipy.special.logsumexp.html>,
# but implementation is easy enough
def _logsumexp(x: ArrayLike):
    c = np.max(x)
    return c + np.log(np.sum(np.exp(x - c)))


def mean(x: ArrayLike, p: float = 1) -> np.ndarray:
    """Generalized mean.

    See <https://github.com/numpy/numpy/issues/19341> for the numpy issue.
    """
    x = np.asarray(x)

    n = len(x)
    if p == 1:
        return np.mean(x)

    if p == -np.inf:
        return np.min(np.abs(x))

    if p == 0:
        # first compute the root, then the product, to avoid numerical
        # difficulties with too small values of prod(x)
        if np.any(x < 0.0):
            msg = "p=0 only works with nonnegative x."
            raise ValueError(msg)
        return np.prod(np.power(x, 1 / n))
        # alternative:
        # return np.exp(np.mean(np.log(x)))

    if p == np.inf:
        return np.max(np.abs(x))

    if np.all(x > 0.0):
        # logsumexp trick to avoid overflow for large p
        # only works for positive x though
        return np.exp((_logsumexp(p * np.log(x)) - np.log(n)) / p)

    if not isinstance(p, (int, np.integer)):
        msg = f"Non-integer p (={p}) only work with nonnegative x."
        raise TypeError(msg)

    return (np.sum(x**p) / n) ** (1.0 / p)



import numpy as np
import perfplot

import npx

rng = np.random.default_rng(0)

m = 100


def setup(n):
    idx = rng.randomint(0, m, n)
    b = rng.random(n)
    return idx, b


def np_add_at(data):
    a = np.zeros(m)
    idx, b = data
    np.add.at(a, idx, b)
    return a


def npx_add_at(data):
    a = np.zeros(m)
    idx, b = data
    npx.add_at(a, idx, b)
    return a


def npx_sum_at(data):
    idx, b = data
    return npx.sum_at(b, idx, minlength=m)


b = perfplot.bench(
    setup=setup,
    kernels=[np_add_at, npx_add_at, npx_sum_at],
    n_range=[2**k for k in range(23)],
)
b.save("perf-add-at.svg")



import numpy as np

import npx


def test_unique_tol():
    a = [0.1, 0.15, 0.7]

    a_unique = npx.unique(a, 2.0e-1)
    print(a_unique)
    assert np.all(a_unique == [0.1, 0.7])

    a_unique, inv = npx.unique(a, 2.0e-1, return_inverse=True)
    assert np.all(a_unique == [0.1, 0.7])
    assert np.all(inv == [0, 0, 1])


def test_unique_edge_case():
    # 1.1 + 2.2 = 3.3000000000000003
    out = npx.unique([1.1 + 2.2, 3.3], tol=0.1)
    assert len(out) == 1



import numpy as np

import npx


def test_1d():
    a = [1, 2, 1]
    a_unique = npx.unique_rows(a)
    assert np.all(a_unique == [1, 2])


def test_2d():
    a = [[1, 2], [1, 4], [1, 2]]
    a_unique = npx.unique_rows(a)
    assert np.all(a_unique == [[1, 2], [1, 4]])


def test_3d():
    # entries are matrices
    # fails for some reason. keep an eye on
    # <https://stackoverflow.com/q/67128631/353337>
    a = [[[3, 4], [-1, 2]], [[3, 4], [-1, 2]]]
    a_unique = npx.unique_rows(a)
    assert np.all(a_unique == [[[3, 4], [-1, 2]]])


def test_return_all():
    a = [[1, 2], [1, 4], [1, 2]]
    a_unique, inv, count = npx.unique_rows(a, return_inverse=True, return_counts=True)
    assert np.all(a_unique == [[1, 2], [1, 4]])
    assert np.all(inv == [0, 1, 0])
    assert np.all(count == [2, 1])


def test_empty():
    # empty mesh
    a = np.empty((1, 0), dtype=int)
    a_unique = npx.unique_rows(a)
    assert np.all(a_unique == [[]])

    a = np.empty((0, 2), dtype=int)
    a_unique = npx.unique_rows(a)
    assert np.all(a_unique == a)


def test_float():
    a = [1.1, 1.2, 1.1]
    out = npx.unique_rows(a)
    ref = np.array([1.2, 1.1])
    assert np.all(np.abs(out - ref) < 1.0e-14)


def test_float_rows():
    a = [[1.1, 0.7], [1.0, 1.2], [1.1, 0.7]]
    out = npx.unique_rows(a)
    ref = [[1.0, 1.2], [1.1, 0.7]]
    assert np.all(np.abs(out - ref) < 1.0e-14)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import ArrayLike


def _unique_tol(
    unique_fun: Callable,
    a: ArrayLike,
    tol: float,
    **kwargs,
) -> np.ndarray | tuple[np.ndarray, ...]:
    a = np.asarray(a)
    # compute 1/tol first. Difference:
    #
    #   int(3.3 / 0.1) = int(32.99999999999999) = 32
    #   int(3.3 * (1.0 / 0.1)) = int(33.0) = 33
    #
    aint = (a * (1.0 / tol)).astype(int)

    return_index = kwargs.pop("return_index", False)

    _, idx, *out = unique_fun(aint, return_index=True, **kwargs)
    unique_a = a[idx]

    if return_index:
        out = [idx, *out]

    if len(out) == 0:
        return unique_a

    return (unique_a, *out)


def unique(
    a: ArrayLike,
    tol: float = 0.0,
    **kwargs,
) -> np.ndarray | tuple[np.ndarray, ...]:
    assert tol >= 0.0
    if tol > 0.0:
        return _unique_tol(np.unique, a, tol, **kwargs)

    return np.unique(a, **kwargs)


def unique_rows(a: ArrayLike, **kwargs) -> np.ndarray | tuple[np.ndarray, ...]:
    # The numpy alternative `np.unique(a, axis=0)` is slow; cf.
    # <https://github.com/numpy/numpy/issues/11136>.
    a = np.asarray(a)

    a_shape = a.shape
    a = a.reshape(a.shape[0], np.prod(a.shape[1:], dtype=int))

    b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    out = np.unique(b, **kwargs)

    # out[0] are the sorted, unique rows
    if isinstance(out, tuple):
        out = list(out)
        out[0] = out[0].view(a.dtype).reshape(out[0].shape[0], *a_shape[1:])
        if len(out) > 1:
            out[1] = out[1].flatten()
        if len(out) > 2:
            out[2] = out[2].flatten()
        out = tuple(out)
    else:
        out = out.view(a.dtype).reshape(out.shape[0], *a_shape[1:])

    return out
